"""Group3 Raise SystemExit('message') in argument_parser()."""
__title__ = __doc__


def argument_parser():
    """As required by spicerack.cookbook."""
    raise SystemExit('argument_parser')


def run(_args, _spicerack):
    """As required by spicerack.cookbook."""
