"""Group3 argparse."""
import argparse


__title__ = __doc__


def argument_parser():
    """As required by spicerack.cookbook."""
    return argparse.ArgumentParser('Argparse')


def run(_args, _spicerack):
    """As required by spicerack.cookbook."""
